package ca.yorku.cse.mack.tilttarget;

import android.app.Activity;
import android.content.pm.ActivityInfo;
import android.content.res.Configuration;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.media.MediaPlayer;
import android.media.MediaScannerConnection;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.os.Environment;
import android.os.SystemClock;
import android.util.Log;
import android.view.Display;
import android.view.Surface;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.TextView;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Locale;
import java.util.Random;
import java.util.Vector;

/**
 * <h1>TiltTarget</h1>
 *
 * <h3>Summary</h3>
 *
 * <ul>
 * <li>Experiment software to investigate device tilt as a target selection method
 * </ul>
 *
 * <h3>Related References</h3>
 *
 * The following publication presents research where this software was used.
 * <p>
 *
 * <ul>
 * <li>
 * <a href="http://www.yorku.ca/mack/bhci2013.html">Exploring tilt-based text input for
 * mobile devices with teenagers</a>, by Fitton, MacKenzie, Read, and Horton (<i>BHCI 2013</i>).
 * <p>
 *
 * <li>The study of extraneous conditions that affect tilt-based pointer movement, by
 * Serackis, Miniotas, Katkevi&ccaron;ius, Krukonis, and Plonis (<i>BIOMDLORE 2016</i>).
 * </ul>
 * <p>
 *
 * <h3>Setup Parameters</h3>
 *
 * The application launches with a setup dialog: (click to enlarge)
 * <p>
 *
 * <center>
 * <a href="TiltTarget-1.jpg"><img src="TiltTarget-1.jpg" width=200 alt="image"></a>
 * </center><p>
 *
 * In all, there are 14 setup parameters, organized in 12 spinners and two checkboxes. The setup parameters are
 * explained as follows. <p>
 *
 * <blockquote>
 * <table border="1" cellspacing="0" cellpadding="6" valign="top">
 *
 * <tr bgcolor="#cccccc">
 * <th align="center">Parameter
 * <th align="center">Description
 *
 * <tr valign="top"> <td>Participant Code <td>A code to identify the user being tested. In experimental research, a user
 * is typically called a "participant". <p>
 *
 * <tr valign="top"> <td>Session Code <td>A code to identify the session. This code could be used in an a longitudinal
 * experiment, where participants are tested over multiple sessions. Each session might involve an hour of testing, with
 * one session per day over multiple days. <p>
 *
 * <tr valign="top">
 * <td>Block Code
 * <td>Note: There is an inactive setup parameter for "Block code". The block code is generated automatically by the
 * software. <p>
 *
 * <tr valign="top"> <td>Group Code <td>A code to identify the group. If the experiment uses a within-subjects design,
 * participants are usually assigned to groups. This code identifies the group. This procedure is known as
 * <i>counterbalancing</i>, which is used to offset order effects. (See section 5.11 in the course's Suggested Readings
 * for complete details.) <p>
 *
 * <tr valign="top"> <td>Condition Code <td>A code to identify an external condition. This setup item is an editable
 * text field. Tap to popup an input method editor to change the value. <p>
 *
 * The condition code is useful if the experiment included an independent variable pertaining to an environmental
 * condition or a context of use. As an example, an experiment might include "mobility" as an independent variable with
 * levels "sitting" and "walking". The code "Sitting" might be used for the sitting condition and the code "Walking" for
 * the walking condition. These codes appear in the output data to facilitate analyses. BTW, an example of an experiment
 * using this independent variable is described by Fitton et al. in <a href="http://www.yorku.ca/mack/bhci2013
 * .html">Exploring Tilt-Based Text Input For Mobile Devices With Teenagers</a>. <p>
 *
 * <tr valign="top">
 * <td>Sequences per Block
 * <td>The number of selection sequences for each block of testing.
 * <p>
 *
 * <tr valign="top">
 * <td>Targets per Sequence
 * <td>The number of targets to select for each sequence.  Targets appears as highlighted keys.  Each target to
 * select is chosen at random from the available targets.
 * <p>
 *
 * <tr valign="top"> <td>Keyboard<td>The keyboard to use.  The following options appear: <p>
 *
 * <ul>
 * <li>3x5 - a portrait arrangement of 15 keys, somewhat like a phone keypad
 * <li>4x10 - a landscape arrangement of 40 keys, somewhat like a qwerty keyboard
 * </ul> <p>
 *
 * <tr valign="top">
 * <td>Order of Control
 * <td>The options are<p>
 *
 * <ul>
 * <li>Velocity &ndash; tilting the device controls the velocity of the virtual ball
 * <li>Position &ndash; tilting the device controls the position of the virtual ball
 * </ul><p>
 *
 * <tr valign="top">
 * <td>Tilt Gain
 * <td>A value from the set { Very low, Low, Medium, High, Very high } roughly corresponding to gain
 * sensitivities ranging for very low to very high.<p>
 *
 * <tr valign="top">
 * <td>Dwell time
 * <td>A value from the set { 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000 }.  The value sets the duration in
 * milliseconds that the ball must remain inside a key to select the key
 * <p>
 *
 * <tr valign="top">
 * <td>Sensor sampling rate
 * <td>A value from the set { Normal, UI, Game, Fastest } to set the sampling rate for the device sensor.  See
 * Demo_Sensors for further details.
 *
 * <tr valign="top">
 * <td>Vibrotactile feedback
 * <td>A checkbox item. If selected, enables a vibrotactile pulse to be emitted on each key selection
 *
 * <tr valign="top">
 * <td>Auditory feedback
 * <td>A checkbox item. If selected, enables an auditory click to be emitted on each key
 * selection
 * </table> </blockquote> <p>
 *
 * <h3>Program Operation</h3>
 * <p>
 *
 * After the setup parameters are chosen, the program is launched by clicking the OK button.  The following screen
 * appears:
 * <p>
 *
 * <center>
 * <a href="TiltTarget-2.jpg"><img src="TiltTarget-2.jpg" width=200 alt="image"></a>
 * </center>
 * <p>
 *
 * To begin selecting targets, the virtual ball must be positioned in the green square for one second.  When this
 * occurs, the program transitions to an experiment screen
 * featuring the selected arrangement of targets.  Figures appear below for the 5x3 (left) and 4x10 (right)
 * arrangements.
 * <p>
 *
 * <center>
 * <a href="TiltTarget-3.jpg"><img src="TiltTarget-3.jpg" width=150 alt="image"></a>
 * <a href="TiltTarget-4.jpg"><img src="TiltTarget-4.jpg" width=300 alt="image"></a>
 * </center>
 * <p>
 *
 * The user selects targets by tilting the device to position the ball on a target and then maintaining the ball
 * inside the target for the chosen dwell time.  If the ball is inside the wrong target, a selection still occurs, but
 * is logged as an error.  After the specified number of selections, results appears:
 * <p>
 *
 * <center>
 * <a href="TiltTarget-5.jpg"><img src="TiltTarget-5.jpg" width=200 alt="image"></a>
 * </center>
 * <p>
 *
 * Testing is organized as trials per sequence and sequences per block.  At the end of a block, the application
 * terminates.
 * <p>
 *
 * <h3>Data Storage</h3>
 *
 * Data are also saved in files for follow-up analyses.  There are two files, an "sd1" file which contains data on a
 * per-selection basis (timestamps + keys), and an "sd2" file which contains summary data for each sequence of input.
 * The data files are
 * saved in
 * the public directory on Android devices in a sub-directory called "TiltTargetData".  Below are example
 * files:
 * <p>
 *
 * <ul>
 * <li>sd1 &ndash; <a href="TiltTarget-P00-5x3-S00-B01-G00-C00-POS-TG40-DT500.sd1">
 * TiltTarget-P00-5x3-S00-B01-G00-C00-POS-TG40-DT500.sd1</a>
 * <li>sd2 &ndash; <a href="TiltTarget-P00-5x3-S00-B01-G00-C00-POS-TG40-DT500.sd2">
 * TiltTarget-P00-5x3-S00-B01-G00-C00-POS-TG40-DT500.sd2</a>
 * </ul>
 * <p>
 *
 * The data in the sd2 files are comma-delimited to facilitate importing into Excel or a statistics
 * application for follow-up analyses.  An example is shown below for the data in the sd2 file above.
 * <p>
 *
 * <center>
 * <a href="TiltTarget-6.jpg"><img src="TiltTarget-6.jpg" width=800 alt="image"></a>
 * </center>
 *
 * @author Scott MacKenzie, 2011-2017
 */
public class TiltTargetActivity extends Activity implements SensorEventListener
{
    final static String MYDEBUG = "MYDEBUG"; // for Log.i messages
    final static String APP = "TiltTarget";

    final static int REFRESH_INTERVAL = 20; // milliseconds (screen refreshes @ 50 Hz)

    // int constants to setup a sensor mode (see Demo_TiltMeter API for discussion)
    final static int ORIENTATION = 0;
    final static int ACCELEROMETER_ONLY = 1;
    final static int ACCELEROMETER_AND_MAGNETIC_FIELD = 2;

    final String WORKING_DIRECTORY = "TiltTargetData";
    final String SD2_HEADER = "App,Participant,Session,Block,Group,Condition,Keyboard,Order_of_control,Tilt_gain," +
            "Dwell_time,Desired_sequence,Actual_sequence,Time(ms),Targets,Hits,Misses,Error_rate(%)," +
            "Actual_ball_distance,Minimum_ball_distance,Movement_efficiency(%)\n";

    final String FIVE_BY_THREE = "5x3";
    final String FOUR_BY_TEN = "4x10";
    final float RADIANS_TO_DEGREES = 57.2957795f;
    int sensorMode;
    int defaultOrientation;
    TextView sequenceEnteredField, progressTextField;
    ExperimentPanel experimentPanel;
    String participantCode, sessionCode, blockCode, groupCode, conditionCode;
    int sequencesPerBlock, targetsPerSequence;
    int sequenceCount;
    int keyToSelect, lastKeyToSelect;
    int numberOfKeys;
    int blockNumber;
    String keyboard;
    String sequence;
    String sd2Leader;
    StringBuilder sequenceEntered, transcribedText, wordStem, resultsString, sd1Stuff, sd2Stuff;
    StringBuilder desiredSequence;
    Random r = new Random();
    Vector<Sample> samples;
    Vector<Point> actualPath, minimumPath;
    long t1, t2;
    int selectionCount; // the number of selections (whether or not the selection was on the target)
    int misses; // number of selections *not* on the desired target

    String orderOfControl;
    int tiltGain;
    int dwellTime;
    int samplingRate;
    boolean vibrotactileFeedback, auditoryFeedback;
    BufferedWriter sd1, sd2; // summary data 1 and summary data 2 files for output data
    File f1, f2;
    Display display;
    ScreenRefreshTimer refreshScreen;

    float[] accValues = new float[3];
    float[] magValues = new float[3];
    float x, y, z, pitch, roll;

    /**
     * Below are the alpha values for the low-pass filter. The four values in each array are for the
     * slowest to fastest sampling rates, respectively. These values were determined by trial and
     * error. There is a trade-off. Generally, lower values produce smooth but sluggish responses,
     * while higher values produced jerky but fast responses. There is also a difference by order of
     * control; hence the use of two arrays. Furthermore, there appears to be difference by device
     * (e.g., Samsung Galaxy Tab 10.1 vs. HTC Desire C). More work is needed here.
     */
    // final float[] ALPHA_VELOCITY = { 0.99f, 0.8f, 0.4f, 0.15f };
    // final float[] ALPHA_POSITION = { 0.5f, 0.3f, 0.15f, 0.10f };
    float alpha;

    private SensorManager sm;
    private Sensor sO, sA, sM;

    // called when the activity is first created.
    @Override
    public void onCreate(Bundle savedInstanceState)
    {
        super.onCreate(savedInstanceState);
        setContentView(R.layout.main);

        sequenceEnteredField = (EditText)findViewById(R.id.presentedTry);
        progressTextField = (EditText)findViewById(R.id.transcribedTry);
        experimentPanel = (ExperimentPanel)findViewById(R.id.experimentpanel);

        display = getWindowManager().getDefaultDisplay();
        defaultOrientation = getDefaultDeviceOrientation();

        // init study parameters
        Bundle b = getIntent().getExtras();
        participantCode = b.getString("participantCode");
        sessionCode = b.getString("sessionCode");
        // blockCode = b.getString("blockCode");
        groupCode = b.getString("groupCode");
        conditionCode = b.getString("conditionCode");
        sequencesPerBlock = b.getInt("sequencesPerBlock");
        targetsPerSequence = b.getInt("targetsPerSequence");
        keyboard = b.getString("keyboard");
        orderOfControl = b.getString("orderOfControl");
        tiltGain = b.getInt("tiltGain");
        dwellTime = b.getInt("dwellTime");
        samplingRate = b.getInt("samplingRate");
        vibrotactileFeedback = b.getBoolean("vibrotactileFeedback");
        auditoryFeedback = b.getBoolean("auditoryFeedback");

        if (keyboard.equals(FIVE_BY_THREE))
        {
            numberOfKeys = 15;
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);
        } else if (keyboard.equals(FOUR_BY_TEN))
        {
            numberOfKeys = 40;
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE);
        }

        sequenceEntered = new StringBuilder(100);
        desiredSequence = new StringBuilder(100);
        transcribedText = new StringBuilder(100);
        progressTextField.setCursorVisible(false);

        samples = new Vector<Sample>();
        actualPath = new Vector<Point>();
        minimumPath = new Vector<Point>();
        wordStem = new StringBuilder(100);
        selectionCount = 0;
        misses = 0;
        resultsString = new StringBuilder(500);
        sequenceCount = 0;
        sequence = "";
        lastKeyToSelect = -1;
        updateProgressText();

        // instantiate the screen refresh timer (but don't start it until window has focus)
        refreshScreen = new ScreenRefreshTimer(REFRESH_INTERVAL, REFRESH_INTERVAL);

        // ===================
        // File initialization
        // ===================

        // make a working directory (if necessary) to store data files
        File dataDirectory = new File(Environment.getExternalStorageDirectory() + File.separator + WORKING_DIRECTORY);
        if (!dataDirectory.exists() && !dataDirectory.mkdirs())
        {
            Log.e(MYDEBUG, "ERROR --> FAILED TO CREATE DIRECTORY: " + WORKING_DIRECTORY);
            super.onDestroy(); // cleanup
            this.finish(); // terminate
        }
        Log.i(MYDEBUG, "Working directory=" + dataDirectory);

        /**
         * The following do-loop creates data files for output and a string sd2Leader to write to the sd2
         * output files.  Both the filenames and the sd2Leader are constructed by combining the setup parameters
         * so that the filenames and sd2Leader are unique and also reveal the conditions used for the block of input.
         *
         * The block code begins "B01" and is incremented on each loop iteration until an available
         * filename is found.  The goal, of course, is to ensure data files are not inadvertently overwritten.
         */
        blockNumber = 0;
        do
        {
            ++blockNumber;
            blockCode = String.format(Locale.CANADA, "B%02d", blockNumber);
            String baseFilename = String.format(Locale.CANADA, "%s-%s-%s-%s-%s-%s-%s-%s-TG%d-DT%d", APP, participantCode, keyboard,
                    sessionCode, blockCode, groupCode, conditionCode, orderOfControl.substring(0, 3).toUpperCase(),
                    tiltGain, dwellTime);
            f1 = new File(dataDirectory, baseFilename + ".sd1");
            f2 = new File(dataDirectory, baseFilename + ".sd2");

            // also make a comma-delimited leader that will begin each data line written to the sd2 file
            sd2Leader = String.format(Locale.CANADA, "%s,%s,%s,%s,%s,%s,%s,%s,TG%d,DT%d,", APP, participantCode, sessionCode,
                    blockCode, groupCode, conditionCode, keyboard, orderOfControl.substring(0, 3)
                            .toUpperCase(),
                    tiltGain, dwellTime);

        } while (f1.exists() || f2.exists());

        try
        {
            sd1 = new BufferedWriter(new FileWriter(f1));
            sd2 = new BufferedWriter(new FileWriter(f2));

            // output header in sd2 file
            sd2.write(SD2_HEADER, 0, SD2_HEADER.length());
            sd2.flush();
        } catch (IOException e)
        {
            Log.e(MYDEBUG, "ERROR OPENING DATA FILES! e=" + e.toString());
            super.onDestroy();
            this.finish();
        } // end file initialization

        // send the setup up parameters to the experiment panel
        experimentPanel.setParameters(keyboard, orderOfControl, tiltGain, dwellTime, vibrotactileFeedback,
                MediaPlayer.create(this, R.raw.click), auditoryFeedback);

        // prevent IME (e.g., a soft keyboard) from popping up
        getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

        alpha = 0.1f; // constant for low-pass filter (might need to fiddle with this)

        // get sensors
        sm = (SensorManager)getSystemService(SENSOR_SERVICE);
        sO = sm.getDefaultSensor(Sensor.TYPE_ORIENTATION); // supported on many devices
        sA = sm.getDefaultSensor(Sensor.TYPE_ACCELEROMETER); // supported on most devices
        sM = sm.getDefaultSensor(Sensor.TYPE_MAGNETIC_FIELD); // null on many devices

        // setup the sensor mode (see API for discussion)
        if (sO != null)
        {
            sensorMode = ORIENTATION;
            sA = null;
            sM = null;
            Log.i(MYDEBUG, "Sensor mode: ORIENTATION");
        } else if (sA != null && sM != null)
        {
            sensorMode = ACCELEROMETER_AND_MAGNETIC_FIELD;
            Log.i(MYDEBUG, "Sensor mode: ACCELEROMETER_AND_MAGNETIC_FIELD");
        } else if (sA != null)
        {
            sensorMode = ACCELEROMETER_ONLY;
            Log.i(MYDEBUG, "Sensor mode: ACCELEROMETER_ONLY");
        } else
        {
            Log.i(MYDEBUG, "Can't run demo.  Requires Orientation sensor or Accelerometer");
            this.finish();
        }

        // NOTE: sensor listeners are registered in onResume

    } // end onCreate

    @Override
    public void onWindowFocusChanged(boolean hasFocus)
    {
        experimentPanel.resultsMode = true;

        // now safe to start the timer which refreshes the UI (updates every REFRESH_INTERVAL  milliseconds)
        refreshScreen.start();
    }

    @Override
    protected void onResume()
    {
        super.onResume();
        if (sO != null)
            sm.registerListener(this, sO, samplingRate);
        else
        {
            sm.registerListener(this, sA, samplingRate);
            sm.registerListener(this, sM, samplingRate);
        }
    }

    @Override
    protected void onPause()
    {
        super.onPause();
        sm.unregisterListener(this);
    }

    /*
     * Cancel the timer when the activity is stopped. If we don't do this, the timer continues after
     * the activity finishes. See Mort's answer in...
     *
     * http://stackoverflow.com/questions/15144232/countdowntimer-continues-to-tick-in-background-how
     * -do-i-retrieve-that-count-in
     */
    @Override
    public void onStop()
    {
        refreshScreen.cancel();
        super.onStop();
    }

    // implement SensorEventListener methods
    @Override
    public void onAccuracyChanged(Sensor sensor, int accuracy)
    {
    }

    @Override
    public void onSensorChanged(SensorEvent se)
    {
        // ===============================
        // DETERMINE DEVICE PITCH AND ROLL
        // ===============================

        switch (sensorMode)
        {
            case ORIENTATION: // ========================================================

				/*
                 * Use this mode if the device has an orientation sensor.
				 */

                if (se.sensor.getType() != Sensor.TYPE_ORIENTATION)
                {
                    Log.i(MYDEBUG, "Return now. Sensor event from " + se.sensor.getName());
                    return;
                }

                // This bit of fiddling is necessary so the app will work on different devices.
                switch (defaultOrientation)
                {
                    case Configuration.ORIENTATION_PORTRAIT:
                    {
                        // e.g., Nexus 4
                        pitch = se.values[1];
                        roll = se.values[2];
                        break;
                    }
                    case Configuration.ORIENTATION_LANDSCAPE:
                    {
                        // e.g., Samsung Galaxy Tab 10.1
                        pitch = se.values[2];
                        roll = -se.values[1];
                        break;
                    }
                }
                break;

            case ACCELEROMETER_AND_MAGNETIC_FIELD: // ===================================

				/*
                 * Use this mode if the device has both an accelerometer and a magnetic field sensor
				 * (but no orientation sensor). See...
				 *
				 * http://blog.thomnichols.org/2012/06/smoothing-sensor-data-part-2
				 */

                // smooth the sensor values using a low-pass filter
                if (se.sensor.getType() == Sensor.TYPE_ACCELEROMETER)
                    accValues = lowPass(se.values.clone(), accValues, alpha);
                if (se.sensor.getType() == Sensor.TYPE_MAGNETIC_FIELD)
                    magValues = lowPass(se.values.clone(), magValues, alpha);

                if (accValues != null && magValues != null)
                {
                    // compute pitch and roll
                    float R[] = new float[9];
                    float I[] = new float[9];
                    boolean success = SensorManager.getRotationMatrix(R, I, accValues, magValues);
                    if (success) // see SensorManager API
                    {
                        float[] orientation = new float[3];
                        SensorManager.getOrientation(R, orientation); // see getOrientation API
                        pitch = orientation[1] * RADIANS_TO_DEGREES;
                        roll = -orientation[2] * RADIANS_TO_DEGREES;
                    }
                }
                break;

            case ACCELEROMETER_ONLY: // =================================================

				/*
                 * Use this mode if the device has an accelerometer but no magnetic field sensor and
				 * no orientation sensor (e.g., HTC Desire C, Asus MeMOPad). This algorithm doesn't
				 * work quite as well, unfortunately. See...
				 *
				 * http://www.hobbytronics.co.uk/accelerometer-info
				 */

                // smooth the sensor values using a low-pass filter
                if (se.sensor.getType() == Sensor.TYPE_ACCELEROMETER)
                    accValues = lowPass(se.values.clone(), accValues, alpha);

                x = accValues[0];
                y = accValues[1];
                z = accValues[2];
                pitch = (float)Math.atan(y / Math.sqrt(x * x + z * z)) * RADIANS_TO_DEGREES;
                roll = (float)Math.atan(x / Math.sqrt(y * y + z * z)) * RADIANS_TO_DEGREES;
                break;
        }

        if (keyboard.equals(FOUR_BY_TEN))
        {
            float temp = pitch;
            pitch = -roll;
            roll = temp;
        }
    }

    /*
     * Get the default orientation of the device. This is needed to correctly map the sensor data
     * for pitch and roll (see onSensorChanged). See...
     *
     * http://stackoverflow.com/questions/4553650/how-to-check-device-natural-default-orientation-on-
     * android-i-e-get-landscape
     */
    public int getDefaultDeviceOrientation()
    {
        WindowManager windowManager = (WindowManager)getSystemService(WINDOW_SERVICE);
        Configuration config = getResources().getConfiguration();
        int rotation = windowManager.getDefaultDisplay().getRotation();

        if (((rotation == Surface.ROTATION_0 || rotation == Surface.ROTATION_180) && config.orientation ==
                Configuration.ORIENTATION_LANDSCAPE)
                || ((rotation == Surface.ROTATION_90 || rotation == Surface.ROTATION_270) && config.orientation ==
                Configuration.ORIENTATION_PORTRAIT))
            return Configuration.ORIENTATION_LANDSCAPE;
        else
            return Configuration.ORIENTATION_PORTRAIT;
    }

    public void doEndOfSequence()
    {
        experimentPanel.resultsMode = true;

        // compute distance the ball moved
        float distanceMoved = 0.0f;
        for (int i = 1; i < actualPath.size(); ++i)
        {
            int x1 = (actualPath.elementAt(i - 1)).x;
            int y1 = (actualPath.elementAt(i - 1)).y;
            int x2 = (actualPath.elementAt(i)).x;
            int y2 = (actualPath.elementAt(i)).y;
            int dx = x2 - x1;
            int dy = y2 - y1;
            distanceMoved += (float)Math.sqrt(dx * dx + dy * dy);
        }

        // compute the minimum possible distance the ball could have moved
        // NOTE: calculation uses the middle of the targets. Actual minimum
        // may be less if the ball is consistently moved only to the nearest edge of the targets.
        float minimumDistance = 0.0f;
        for (int i = 1; i < minimumPath.size(); ++i)
        {
            int x1 = (minimumPath.elementAt(i - 1)).x;
            int y1 = (minimumPath.elementAt(i - 1)).y;
            int x2 = (minimumPath.elementAt(i)).x;
            int y2 = (minimumPath.elementAt(i)).y;
            int dx = x2 - x1;
            int dy = y2 - y1;
            minimumDistance += (float)Math.sqrt(dx * dx + dy * dy);
        }

        // =========================
        // Results presented to user
        // =========================

        resultsString.delete(0, resultsString.length());
        resultsString.append("Thank you!\n \n");

        String s = String.format(Locale.CANADA, "Block: %d\n", blockNumber);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Sequence: %d of %d\n", (sequenceCount + 1), sequencesPerBlock);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Total time: %.2f s\n", t2 / 1000.0f);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Selections: %d\n", selectionCount);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Misses: %d\n", misses);
        resultsString.append(s);

        float errorRate = (float)misses / selectionCount * 100.0f;
        s = String.format(Locale.CANADA, "Error rate: %.1f%%\n", errorRate);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Ball distance: %d px\n", (int)distanceMoved);
        resultsString.append(s);

        s = String.format(Locale.CANADA, "Min distance: %d px\n", (int)minimumDistance);
        resultsString.append(s);

        /**
         * NOTE: Movement efficiency can exceed 100% if the ball movement is direct AND selection occurs at
         * the nearest edge of each target.  It can also exceed 100% due to errors (i.e., selections *not* on the
         * desired target).  If analysing movement efficiency as a dependent variable, probably only error-free
         * sequences should be analysed.
         */
        float movementEfficiency = minimumDistance / distanceMoved * 100.0f;
        s = String.format(Locale.CANADA, "Movement efficiency: %.1f%%\n\n", movementEfficiency);
        resultsString.append(s);

        resultsString.append("GREEN to continue");
        experimentPanel.infoString = resultsString;

        // ==================================
        // Results saved to output data files
        // ==================================

        // sd2 stuff first

        sd2Stuff = new StringBuilder(500);
        sd2Stuff.append(sd2Leader);
        s = String.format(Locale.CANADA, "%s,%s,%d,%d,%d,%d,%.2f,%d,%d,%.1f\n", desiredSequence, sequenceEntered, t2,
                selectionCount, (selectionCount - misses), misses, errorRate, (int)distanceMoved, (int)minimumDistance,
                movementEfficiency);
        sd2Stuff.append(s);

        // sd1 stuff next

        sd1Stuff = new StringBuilder(100);
        sd1Stuff.append(String.format("%s\n", desiredSequence));
        sd1Stuff.append(String.format("%s\n", sequenceEntered));
        for (int i = 0; i < samples.size(); ++i)
            sd1Stuff.append(String.format("%s\n", samples.elementAt(i).toString()));
        sd1Stuff.append("-----\n");

        // write to data files
        try
        {
            sd1.write(sd1Stuff.toString(), 0, sd1Stuff.length());
            sd1.flush();
            sd2.write(sd2Stuff.toString(), 0, sd2Stuff.length());
            sd2.flush();
        } catch (IOException e)
        {
            System.err.println("ERROR WRITING TO DATA FILE!\n" + e);
            System.exit(1);
        }

        // check if last sequence in block
        ++sequenceCount;
        if (sequenceCount == sequencesPerBlock)
            experimentPanel.done = true; // ... but display results from last sequence before quitting
        else
            doNewSequence();
    } // end doEndOfSequence

    public void doNewSequence()
    {
        sequenceEntered.replace(0, sequenceEntered.length(), "");
        sequenceEnteredField.setText(sequenceEntered);
        desiredSequence.replace(0, desiredSequence.length(), "");

        t1 = SystemClock.elapsedRealtime();
        selectionCount = 0;
        misses = 0;
        updateProgressText();

        pickNextTargetKey();

        experimentPanel.readyToDraw = true;

        samples = new Vector<Sample>();
        actualPath = new Vector<Point>();
        minimumPath = new Vector<Point>();
    }

    public void pickNextTargetKey()
    {
        // pick the target to select
        do
        {
            keyToSelect = r.nextInt(numberOfKeys);
        } while (keyToSelect == lastKeyToSelect);
        lastKeyToSelect = keyToSelect;

        experimentPanel.keyIndexToSelect = keyToSelect;
        desiredSequence.append(experimentPanel.KEY_LABELS.charAt(keyToSelect));
        minimumPath.addElement(new Point((int)(experimentPanel.key[keyToSelect].x), (int)(experimentPanel
                .key[keyToSelect].y)));
    }

    public void updateProgressText()
    {
        StringBuilder s = new StringBuilder("Targets:     Hits:     Misses:       ");

        // Targets:     Hits:     Misses:
        // 012345678901234567890123456789012345678901234567890123
        //           1         2         3         4         5
        // indices determined from the ruler above
        s.replace(9, 11, "" + targetsPerSequence);
        s.replace(19, 20, "" + (selectionCount - misses));
        s.replace(31, 33, "" + misses);
        progressTextField.setText(s.toString());
    }

    // Done! close data files and exit
    private void doDone()
    {
        try
        {
            sd1.close();
            sd2.close();

            // Make the saved data files visible in Windows Explorer
            // There seems to be bug doing this with Android 4.4. I'm using the following
            // code, instead of sendBroadcast. See...
            // http://code.google.com/p/android/issues/detail?id=38282
            MediaScannerConnection.scanFile(this, new String[] {f1.getAbsolutePath(), f2.getAbsolutePath()}, null,
                    null);
        } catch (IOException e)
        {
            // work on later (but no console to write to)
        }
        refreshScreen.cancel(); // cancel timer (no more UI updates)
        finish();
    }

    // Low pass filter (smoothing algorithm) for sensor data.
    // See http://blog.thomnichols.org/2011/08/smoothing-sensor-data-with-a-low-pass-filter
    // More work is needed to decide on the best alpha. See comments above.
    protected float[] lowPass(float[] input, float[] output, float alpha)
    {
        for (int i = 0; i < input.length; i++)
            output[i] = output[i] + alpha * (input[i] - output[i]);
        return output;
    }

    // =============================================
    // Sample - simple class to hold a key-time data
    // =============================================

    private class Sample
    {
        private long time;
        private String key;

        Sample(long timeArg, String keyArg)
        {
            time = timeArg;
            key = keyArg;
        }

        public String toString()
        {
            return time + ", " + key;
        }
    }

    // =========================================
    // Point - simple class to hold an x-y point
    // =========================================

    private class Point
    {
        int x, y;

        Point(int xArg, int yArg)
        {
            x = xArg;
            y = yArg;
        }
    }

    // screen updates are initiated in onFinish which executes every REFRESH_INTERVAL milliseconds
    public class ScreenRefreshTimer extends CountDownTimer
    {
        ScreenRefreshTimer(long millisInFuture, long countDownInterval)
        {
            super(millisInFuture, countDownInterval);
        }

        @Override
        public void onTick(long millisUntilFinished)
        {
        }

        @Override
        public void onFinish()
        {
            //Log.i(MYDEBUG, "onFinish!");
            float tiltMagnitude = (float)Math.sqrt(pitch * pitch + roll * roll);
            float tiltAngle = tiltMagnitude == 0f ? 0f : (float)Math.asin(roll / tiltMagnitude) * RADIANS_TO_DEGREES;

            if (pitch > 0 && roll > 0)
                tiltAngle = 360f - tiltAngle;
            else if (pitch > 0 && roll < 0)
                tiltAngle = -tiltAngle;
            else if (pitch < 0 && roll > 0)
                tiltAngle = tiltAngle + 180f;
            else if (pitch < 0 && roll < 0)
                tiltAngle = tiltAngle + 180f;

            experimentPanel.updateTilt(tiltAngle, tiltMagnitude);

            // record the path of the ball (make sure the first target has been selected)
            if (selectionCount > 0)
                actualPath.addElement(new Point((int)experimentPanel.xBallCenter, (int)experimentPanel.yBallCenter));

            // key selected stuff done here (perhaps move to separate method)
            if (experimentPanel.keySelected)
            {
                if (experimentPanel.reallyDone) // green key selected after viewing results after last sequence
                {
                    doDone(); // close data files and call finish
                } else if (experimentPanel.beginSequence)  // green key selected to begin the sequence
                {
                    experimentPanel.beginSequence = false;
                    doNewSequence();

                } else // this is a key selection "within the sequence"
                {
                    ++selectionCount;
                    t2 = SystemClock.elapsedRealtime() - t1;

                    Key k = experimentPanel.tentativeKey;

                    // if the key selected was *not* the target key, increment misses
                    if (!experimentPanel.targetKeySelected)
                        ++misses;

                    // update the sequence entered field in the UI
                    sequenceEntered.append(k.label);
                    sequenceEnteredField.setText(sequenceEntered);

                    // update the progress field in the UI
                    updateProgressText();

                    // save time stamp and key label
                    samples.addElement(new Sample(t2, k.label));

                    // pick the next target key to select
                    pickNextTargetKey();

                    // clean up and save data, if we're at the end of a sequence
                    if (selectionCount == targetsPerSequence)
                        doEndOfSequence();
                }
                experimentPanel.keySelected = false;
                experimentPanel.targetKeySelected = false;
            }
            experimentPanel.invalidate();
            this.start();
        }
    }
}